﻿Imports System
Imports System.IO
Imports System.Security.Cryptography
Imports System.Security.Cryptography.X509Certificates
Imports System.Security.Cryptography.Xml
Imports System.Text
Imports System.Xml
Imports System.Windows.Forms

Module Helpers

    Function GetToken(ByVal Element As XmlElement, ByVal context As String) As XmlDocument
        ' Vind signedData of signedDataPrescription element en retourneer token al een XmlDocument
        Dim nodeList As XmlNodeList = Element.GetElementsByTagName(context)
        If nodeList.Count <> 1 Then
            Dim aoTokenCountException As New CryptographicException("Geen of meer dan een te ondertekenen token")
            Throw aoTokenCountException
        End If

        Dim doc As New XmlDocument
        doc.PreserveWhitespace = True

        ' Voeg DTD toe omdat anders wsu:Id niet herkend wordt
        AddDtd(doc, context)

        ' Geef token terug als XmlDocument
        doc.LoadXml(nodeList(0).OuterXml)
        Return doc
    End Function

    Function GetTokenDigest(ByVal Filename As String, ByVal ShaVersion As String, ByVal context As String) As String
        ' Geef de digest van het token terug
        Dim doc As New XmlDocument()
        doc.PreserveWhitespace = True

        Dim token As New XmlDocument()
        token.PreserveWhitespace = True

        ' Load the passed XML file using its name.
        doc.Load(Filename)
        token = GetToken(doc.DocumentElement, context)

        ' Zorg voor EXC14N canonicalisatie
        Dim excc14n As New XmlDsigExcC14NTransform
        excc14n.LoadInput(token)
        ' Gebruik juiste hashalgoritme
        Dim sha As HashAlgorithm
        If ShaVersion = "SHA1" Then
            sha = New SHA1CryptoServiceProvider()
        ElseIf ShaVersion = "SHA256" Then
            sha = New SHA256CryptoServiceProvider()
        Else
            Dim aoHashAlgorithmNotSupported As New CryptographicException("Hash algorithme wordt niet ondersteund")
            Throw aoHashAlgorithmNotSupported
        End If

        ' Geef hash als base64 string terug
        Return Convert.ToBase64String(excc14n.GetDigestedOutput(sha))
    End Function

    Sub AddDtd(ByRef doc As XmlDocument, ByVal context As String)
        ' URI reference moet een URI of een geldig XML ID zijn
        ' signedXml class herkent wsu:Id niet als geldig XML ID, wel als deze als zodanig gedeclareerd wordt
        ' DTD stub toevoegen die alleen het wsu:Id declareert als XML ID
        ' Input:
        ' - XML document
        ' - context (signedData voor TA, signedDataPrescription voor EH)
        If Not (doc.DocumentType Is Nothing) Then
            RemoveDtdDeclaration(doc)
        End If

        Dim dtd As String
        dtd = "<!ELEMENT signedData ANY> <!ATTLIST signedData wsu:Id ID #IMPLIED >"
        ' Pas eventueel aan op signedDataPrescription of ander top element in token
        dtd = dtd.Replace("signedData", context)

        Dim doctype As XmlDocumentType
        doctype = doc.CreateDocumentType(context, Nothing, Nothing, dtd)
        doc.PrependChild(doctype)
    End Sub

    Function RemoveDtdDeclaration(ByVal doc As XmlDocument) As XmlDocument
        ' Eventuele DTD of declaration verwijderen
        If TypeOf doc.FirstChild Is XmlDeclaration Then
            doc.RemoveChild(doc.FirstChild)
        End If
        If TypeOf doc.FirstChild Is XmlDocumentType Then
            doc.RemoveChild(doc.FirstChild)
        End If
        Return doc
    End Function

    Function GetXmlNamespaceManager() As XmlNamespaceManager
        ' Namespacemanager met SOAP, WSS, XML DSIG namespaces
        Dim tbl As New NameTable
        Dim nsmgr As New XmlNamespaceManager(tbl)

        With nsmgr
            .AddNamespace("soap", "http://schemas.xmlsoap.org/soap/envelope/")
            .AddNamespace("wsu", "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd")
            .AddNamespace("wss", "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd")
            .AddNamespace("ds", "http://www.w3.org/2000/09/xmldsig#")
            .AddNamespace("hl7", "urn:hl7-org:v3")
            .AddNamespace("ao", "http://www.aortarelease.nl/805/")
        End With
        Return nsmgr
    End Function

    Function GetWsuId(ByVal Element As XmlElement, ByVal context As String, ByVal nsmgr As XmlNamespaceManager) As String
        ' Zoek het wsu:Id op
        ' Input:
        ' - te doorzoeken XML element
        ' - context (signedData of signedDataPrescription)
        ' - namespaces
        ' Output:
        ' - waarde van wsu:id attribuut als string
        Dim nodeList As XmlNodeList

        nodeList = Element.SelectNodes("//ao:" & context & "/@wsu:Id", nsmgr)
        If nodeList.Count = 1 Then
            GetWsuId = nodeList(0).Value
        Else
            Dim aoWsuIdMissing As New CryptographicException("Geen wsu:Id gevonden")
            Throw aoWsuIdMissing
        End If

    End Function

    Function GetStreamAsByteArray(ByVal stream As System.IO.Stream) As Byte()
        ' Helper om stream naar bytes te converteren
        Dim streamLength As Integer = Convert.ToInt32(stream.Length)

        Dim fileData As Byte() = New Byte(streamLength) {}

        stream.Read(fileData, 0, streamLength)
        stream.Close()

        Return fileData
    End Function

    Sub PrintByteArray(ByVal array() As Byte)
        ' Helper bij debuggen, print een btye array in console
        Dim i As Integer
        For i = 0 To array.Length - 1
            Console.Write(String.Format("{0:X}", array(i)))
            If i Mod 4 = 3 Then
                Console.Write(" ")
            End If
        Next i
        Console.WriteLine()
    End Sub 'PrintByteArray

    Function AskForFile(ByVal prompt As String) As String
        ' Helper om een bestandsnaam te kiezen
        ' Input: 
        ' - prompt tekst
        ' Output:
        ' - naam van gekozen bestand (bestand wordt niet geopend)
        Dim ofd As New OpenFileDialog

        ofd.Title = prompt
        ofd.ShowDialog()
        AskForFile = ofd.FileName
    End Function

End Module
