﻿Imports System
Imports System.IO
Imports System.Security.Cryptography
Imports System.Security.Cryptography.X509Certificates
Imports System.Security.Cryptography.Xml
Imports System.Text
Imports System.Xml

Module Verify
    Function VerifyXmlFileWithCert(ByVal SignatureFileName As String, ByVal Cert As X509Certificate2, ByVal nsmgr As XmlNamespaceManager, ByVal context As String) As [Boolean]
        ' Verifieer een XML file tegen een meegezonden certificaat
        ' Input:
        ' - naam XML file
        ' - certificaat waartegen XML zou moeten valideren
        ' - namespacemanager
        ' - context: (signedData voor TA, signedDataPrescription voor EH)
        ' Output:
        ' - valideert ja/nee

        ' Maak een SignedXml object, laad XML
        Dim signedXml As New SignedXml()
        signedXml = GetSignedXml(SignatureFileName, nsmgr, context)

        ' Controleer de digest
        Dim digest As String
        ' Kijk welke versie van SHA gebruikt wordt in bestand
        Dim shaVersion As String = ""
        If signedXml.SignedInfo.References(0).DigestMethod = "http://www.w3.org/2000/09/xmldsig#sha1" Then
            shaVersion = "SHA1"
        ElseIf signedXml.SignedInfo.References(0).DigestMethod = "http://www.w3.org/2001/04/xmlenc#sha256" Then
            shaVersion = "SHA256"
        Else
            Console.WriteLine("Hash algorithme wordt niet ondersteund")
        End If

        ' Controleer de digest van ondertekend token tegen digest in SignedInfo
        ' De CheckSignature functie van SignedXml doet dit ook, voor SHA1 is het dus niet nodig
        ' Omdat SignedXml geen SHA256 ondersteunt, moet dit in code gebeuren
        If shaVersion <> "SHA1" Then
            digest = GetTokenDigest(SignatureFileName, shaVersion, context)
            If Convert.ToBase64String(signedXml.SignedInfo.References(0).DigestValue) <> digest Then
                Dim aoTokenMismatch As New CryptographicException("Token matcht niet met Signature digest")
                Throw aoTokenMismatch
            End If
        End If

        ' Controleer of signature valide is
        Return VerifyXml(signedXml, Cert, nsmgr, context)
    End Function

    Function GetSignedXml(ByVal SignatureFileName As String, ByVal nsmgr As XmlNamespaceManager, ByVal context As String) As SignedXml
        ' Maak van een XML file tegen een SignedXml
        ' Input:
        ' - naam XML file
        ' - namespacemanager
        ' - context: (signedData voor TA, signedDataPrescription voor EH)
        ' Output:
        ' - SignedXml object

        ' Maak een XML doc
        Dim xmlDoc As New XmlDocument()
        ' Altijd PreserveWhitespace aan, anders gaat .Net pretty-printen en valideert sig niet meer
        xmlDoc.PreserveWhitespace = True
        ' DTD tijdelijk toevoegen zodat wsu:Id herkend wordt als XML ID
        AddDtd(xmlDoc, context)
        ' Laad de file
        xmlDoc.Load(SignatureFileName)

        ' Maak een SignedXml object, laad XML
        Dim signedXml As New SignedXml(xmlDoc)

        ' Zoek Signature node
        Dim nodeList As XmlNodeList = xmlDoc.GetElementsByTagName("Signature")
        If nodeList.Count = 0 Then
            Dim aoNoSignature As New CryptographicException("Geen signatures in bestand")
            Throw aoNoSignature
        End If

        ' Signature laden
        ' Code gaat uit van een Signature in het bestand
        ' Wanneer er meerdere Signatures in bestand zitten (b.v. TA en EH token)
        ' dient de code aangepast te worden om de jusite te kiezen.
        signedXml.LoadXml(CType(nodeList(0), XmlElement))

        Return signedXml
    End Function

    Function VerifyXml(ByVal signedXml As SignedXml, ByVal Cert As X509Certificate2, ByVal nsmgr As XmlNamespaceManager, ByVal context As String) As [Boolean]
        ' Verifieer een XML file tegen een meegezonden certificaat
        ' Input:
        ' - SignedXml
        ' - certificaat waartegen XML zou moeten valideren
        ' - namespacemanager
        ' - context: (signedData voor TA, signedDataPrescription voor EH)
        ' Output:
        ' - valideert ja/nee

        ' Test of SHA256 sig gebruikt wordt
        If signedXml.SignatureMethod = "http://www.w3.org/2001/04/xmldsig-more#rsa-sha256" Then
            ' SignedXml class ondersteunt geen SHA256, dus verify op lager niveau
            Return VerifySignature(signedXml, Cert, nsmgr, "SHA256")
        Else
            ' Verify bij SHA1
            Return signedXml.CheckSignature(Cert, False)
        End If
    End Function

    Function VerifySignature(ByVal signedXml As SignedXml, ByVal Cert As X509Certificate2, ByVal nsmgr As XmlNamespaceManager, ByVal ShaVersion As String) As Boolean
        ' Valideert een RSA+SHA256 handtekening op een SignedXml
        ' SignedXml class zelf ondersteunt geen RSA+SHA256
        ' Input:
        ' - te valideren SignedXml
        ' - gebruikt certificaat
        ' - namespacemanager
        ' - shaVersion, is getest en werkt met "SHA256", werkt mogelijk met andere hashes
        ' Output:
        ' - valideert ja/nee
        Dim excc14n As New XmlDsigExcC14NTransform
        Dim si As New XmlDocument()
        Dim xmlString As String
        Dim sigBytes() As Byte
        Dim signedData() As Byte
        Dim key As New RSACryptoServiceProvider
        key = Cert.PublicKey.Key

        ' Altijd PreserveWhitespace aan, anders gaat .Net pretty-printen en valideert sig niet meer
        si.PreserveWhitespace = True
        ' Neem SignedInfo deel van SignedXml als string voor vervangen URI
        xmlString = signedXml.SignedInfo.GetXml.OuterXml()
        ' Laden als XML
        si.LoadXml(xmlString)
        ' Exclusive XML Canonicalization
        excc14n.LoadInput(si)
        signedData = excc14n.GetDigestedOutput(New SHA256Managed)
        ' Zet de handtekening binair
        sigBytes = signedXml.SignatureValue

        ' Converteer sig naar Base64 en retourneer
        Return key.VerifyHash(signedData, CryptoConfig.MapNameToOID(ShaVersion), sigBytes)
    End Function
End Module
