package nl.nictiz.digsig;

import java.io.UnsupportedEncodingException;
import java.security.InvalidKeyException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.Signature;
import java.security.SignatureException;

// XXX Should use a proper public base64 library, like apache's commons codecs
import sun.misc.BASE64Encoder;

public abstract class StringBasedSignature {

	/**
	 * Names of algorithms etc...
	 */
	private static final String SIGNING_ALGORITHM_SHA1 = "SHA1withRSA";
	private static final String DIGEST_ALGORTIHM_SHA1 = "SHA1";
	private static final String SIGNING_ALGORITHM_SHA256 = "SHA256withRSA";
	private static final String DIGEST_ALGORTIHM_SHA256 = "SHA-256";
	private static final String UTF8_ENCODING = "UTF-8";

	/**
	 * Constants for the namespaces used.
	 */
	protected static final String WSU_NAMESPACE =
		"http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd";
	protected static final String XML_DSIG_NAMESPACE =
		"http://www.w3.org/2000/09/xmldsig#";

	/**
	 * A base64 encoder.
	 */
	private final static BASE64Encoder encoder = new BASE64Encoder();

	/**
	 * Calculate the digest of the token
	 * @return the base64 encoded digest of the token
	 * @throws UnsupportedEncodingException
	 * @throws NoSuchAlgorithmException
	 */
	protected static String digestToken(String canonicalToken, boolean sha256mode)
		throws UnsupportedEncodingException, NoSuchAlgorithmException
	{
		// The bytes of the token will be digested for in the SignedInfo.
		// Always use the UTF-8 encoding, as per the specification of 
		// XML-Signature canonicalization.
		byte[] tokenBytes = canonicalToken.getBytes(UTF8_ENCODING);

		String digestAlgortihm;
		if (sha256mode) {
			digestAlgortihm = DIGEST_ALGORTIHM_SHA256;
		}
		else {
			digestAlgortihm = DIGEST_ALGORTIHM_SHA1;
		}

		// Calculate the digest over the canonical token
		MessageDigest digester = MessageDigest.getInstance(digestAlgortihm);
		byte[] digest = digester.digest(tokenBytes);

		// The digest will be used in base64
		String encodedDigest = encoder.encode(digest);
//		System.err.println("DigestValue:\t" + encodedDigest);

		return encodedDigest;
	}
	
	/**
	 * Construct the SignedInfo.
	 * @param wssTokenId the reference to the "wsu:Id" in the token
	 * @param encodedDigest the base64 encoded digest of the token
	 * @return a String with the <em>canonical</em> SignedInfo
	 */
	protected static String createSignedInfo(String wssTokenId,
			String encodedDigest, boolean sha256mode)
	{
		String digestURI;
		String signatureURI;

		if (sha256mode) {
			// SHA256 variants
			digestURI = "http://www.w3.org/2001/04/xmlenc#sha256"; 
			signatureURI = "http://www.w3.org/2001/04/xmldsig-more#rsa-sha256";
		}
		else {
			// SHA1 variants
			digestURI = "http://www.w3.org/2000/09/xmldsig#sha1";
			signatureURI = "http://www.w3.org/2000/09/xmldsig#rsa-sha1";
		}

		// NOTE: this String is also canonical, mind whitespace/newlines!
		String canonicalizedSignedInfo =
			"<SignedInfo xmlns=\"http://www.w3.org/2000/09/xmldsig#\">"
			+ "<CanonicalizationMethod Algorithm=\"http://www.w3.org/2001/10/xml-exc-c14n#\">"
			+ "</CanonicalizationMethod>"
			+ "<SignatureMethod Algorithm=\"" + signatureURI + "\">"
			+ "</SignatureMethod><Reference URI=\"#" + wssTokenId + "\">"
			+ "<Transforms>"
			+ "<Transform Algorithm=\"http://www.w3.org/2001/10/xml-exc-c14n#\">"
			+ "</Transform>"
			+ "</Transforms>"
			+ "<DigestMethod Algorithm=\"" + digestURI + "\">"
			+ "</DigestMethod>"
			+ "<DigestValue>" + encodedDigest + "</DigestValue>"
			+ "</Reference>"
			+ "</SignedInfo>";
		return canonicalizedSignedInfo;
	}

	/**
	 * Create the signature over the SignedInfo.
	 * @param key the private key to sign with
	 * @param canonicalizedSignedInfo the SignedInfo in canonical format
	 * @return the base64 encoded signature of the SignedInfo
	 * @throws UnsupportedEncodingException
	 * @throws NoSuchAlgorithmException
	 * @throws InvalidKeyException
	 * @throws SignatureException
	 */
	protected static String signSignedInfo(PrivateKey key,
			String canonicalizedSignedInfo, boolean sha256mode)
		throws UnsupportedEncodingException, NoSuchAlgorithmException,
			InvalidKeyException, SignatureException
	{
		// The bytes of the signedInfo will be signed. Always use the UTF-8
		// encoding, as per the specification of XML-Signature canonicalization.
		byte[] signedInfoBytes =
			canonicalizedSignedInfo.getBytes(UTF8_ENCODING);

		String signingAlgorithm;
		if (sha256mode) {
			signingAlgorithm = SIGNING_ALGORITHM_SHA256;
		}
		else {
			signingAlgorithm = SIGNING_ALGORITHM_SHA1;
		}

		// Sign the SingedInfo.
		Signature signer = Signature.getInstance(signingAlgorithm);
		signer.initSign(key);
		signer.update(signedInfoBytes);
		byte[] signature = signer.sign();

		// This should be base64'd as well
		String encodedSignature = encoder.encode(signature);
//		System.err.println("Signature:\t" + encodedSignature);
		return encodedSignature;
	}

	protected static String base64encode(byte[] bytes) {
		return encoder.encode(bytes);
	}

}
