package nl.nictiz.digsig;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;

import javax.xml.XMLConstants;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;

import org.w3c.dom.Document;
import org.xml.sax.SAXException;

/**
 * Some common methods for reading and showing XML Documents in Java.
 */
public class XmlCommon {

	private static final String WSU_SCHEMA_FILENAME = "wsu.xsd";

	/**
	 * Load the XML document as a DOM Document from a file
	 * @param fileName the file with the XML document
	 * @return
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 * @throws IOException
	 * @throws FileNotFoundException
	 */
	public static Document getDocument(String fileName)
		throws ParserConfigurationException, SAXException, IOException,
			FileNotFoundException
	{

		// Java XML processing...
		DocumentBuilderFactory documentBuilderFactory =
			DocumentBuilderFactory.newInstance();

		// Java 5 will recognize a wsu:Id as an ID. The XML parser in java 6 is
		// more strict though, and will not recognize wsu:Id without the schema.
		// Loading the schema will break on Java 5 however, since only the
		// schema part for "wsu" is loaded. Therefore, a bit of version specific
		// code...
		if (!System.getProperty("java.version").startsWith("1.5")) {
			// Explicitly load the wsu schema. Otherwise the "wsu:Id" cannot be
			// used to create the proper reference for the signature.
			File schemaFile = new File(WSU_SCHEMA_FILENAME);
			SchemaFactory schemaFactory =
				SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);
			Schema schema = schemaFactory.newSchema(schemaFile);
			documentBuilderFactory.setSchema(schema);
		}

		documentBuilderFactory.setNamespaceAware(true);
		DocumentBuilder documentBuilder =
			documentBuilderFactory.newDocumentBuilder();
		// You may suppress errors in parsing the XML, caused by the usage of
		// incomplete schema's. This is not a good practice, but the unknown
		// schema warnings may cause some unnecessary distraction
//		documentBuilder.setErrorHandler(null);

		// Read the document
		FileInputStream documentFile = new FileInputStream(fileName);
		Document document = documentBuilder.parse(documentFile);
		return document;
	}

	/**
	 * Show a document to the user on stdout
	 * @param document the Document to show
	 * @throws TransformerFactoryConfigurationError
	 * @throws TransformerConfigurationException
	 * @throws TransformerException
	 */
	public static void displayXml(Document document)
		throws TransformerFactoryConfigurationError,
			TransformerConfigurationException, TransformerException
	{
		// Default java serialization to standard out,
		// uses the XSLT identity transformation
		TransformerFactory transformerFactory =
			TransformerFactory.newInstance();
		Transformer transformer = transformerFactory.newTransformer();
		DOMSource source = new DOMSource(document);
		StreamResult result = new StreamResult(System.out);
		transformer.transform(source, result);
	}

	/**
	 * Apply a stylesheet transformation to a document.
	 * @param document the Document to transform
	 * @param xsltFile the name of the stylesheet file to apply
	 * @return a Document with the result of the transformation
	 * @throws IOException 
	 * @throws SAXException 
	 * @throws ParserConfigurationException 
	 * @throws FileNotFoundException 
	 * @throws TransformerException 
	 */
	public static Document transform(Document document, String xsltFile)
		throws FileNotFoundException, ParserConfigurationException,
			SAXException, IOException, TransformerException
	{
		// Load the stylesheet itself as a XML document
		Document xsltDocument = getDocument(xsltFile);

		// Initialize the XSLT transformation 
		TransformerFactory transformerFactory =
			TransformerFactory.newInstance();
		DOMSource xsltSource = new DOMSource(xsltDocument);
		Transformer transformer = transformerFactory.newTransformer(xsltSource);

		// Create an empty document to capture the result
		DOMSource documentSource = new DOMSource(document);
		DocumentBuilder documentBuilder =
			DocumentBuilderFactory.newInstance().newDocumentBuilder();
		Document resultDocument = documentBuilder.newDocument();

		// Create the resulting document
		DOMResult result = new DOMResult(resultDocument);
		transformer.transform(documentSource, result);
		return resultDocument;

	}

}
