package nl.nictiz.digsig;

import java.io.IOException;
import java.io.InputStream;
import java.security.InvalidAlgorithmParameterException;
import java.security.NoSuchAlgorithmException;
import java.security.Security;
import java.util.Collections;

import javax.xml.crypto.dsig.CanonicalizationMethod;
import javax.xml.crypto.dsig.DigestMethod;
import javax.xml.crypto.dsig.Reference;
import javax.xml.crypto.dsig.SignatureMethod;
import javax.xml.crypto.dsig.SignedInfo;
import javax.xml.crypto.dsig.Transform;
import javax.xml.crypto.dsig.XMLSignatureFactory;
import javax.xml.crypto.dsig.spec.C14NMethodParameterSpec;
import javax.xml.crypto.dsig.spec.TransformParameterSpec;

public abstract class XmlLibrarySignature {

	static {
		// NOTE: In java 5, there is no javax.xml.crypto available by default.
		// Hence, we need to load Cryptographic provider from the xml-security
		// library.
		if (System.getProperty("java.version").startsWith("1.5")) {
			Security.addProvider(new org.jcp.xml.dsig.internal.dom.XMLDSigRI());
		}
	}

	/**
	 * Show the canonical form of the SignedInfo on standard error
	 * @param signedInfo
	 * @throws IOException
	 */
	protected static void showCanonicalizedForm(SignedInfo signedInfo)
			throws IOException {
		InputStream canonicalizedData = signedInfo.getCanonicalizedData();
		int byteCount = canonicalizedData.available();
		byte[] bytes = new byte[byteCount];
		canonicalizedData.read(bytes);
		System.err.println(new String(bytes));
//		TransformerFactory factory = TransformerFactory.newInstance();
//		Transformer newTransformer = factory.newTransformer();
////		ByteArrayOutputStream baos = new ByteArrayOutputStream();
//		newTransformer.transform(new StreamSource(canonicalizedData), new StreamResult(System.err));
//		ByteArrayInputStream bais = new ByteArrayInputStream(baos.toByteArray());
//		Data data = new OctetStreamData(bais);
//		CanonicalizationMethod newCanonicalizationMethod = signatureFactory.newCanonicalizationMethod(CanonicalizationMethod.EXCLUSIVE, (C14NMethodParameterSpec)null);
//		newCanonicalizationMethod.transform(signedInfo, null);
	}

	/**
	 * Construct the signed information block for the signature. Contains
	 *   signing methods and a reference to the data to sign
	 * @param signatureFactory
	 * @param nodeId the ID of the node to sign
	 * @return
	 * @throws NoSuchAlgorithmException
	 * @throws InvalidAlgorithmParameterException
	 */
	protected static SignedInfo makeSignedInfo(
			XMLSignatureFactory signatureFactory, String nodeId,
			boolean sha256mode)
			throws NoSuchAlgorithmException, InvalidAlgorithmParameterException
	{
		
		// The reference to the data to sign
		Reference reference =
			makeReferenceById(signatureFactory, nodeId, sha256mode);

		String signatureAlgorithm = null;
		if (sha256mode) {
			// The javax.xml.crypto API does not have a constant for RSA-SHA256
			signatureAlgorithm = "http://www.w3.org/2001/04/xmldsig-more#rsa-sha256";
		}
		else {
			signatureAlgorithm = SignatureMethod.RSA_SHA1;
		}
		
		// Get the methods for the actual signature itself
		CanonicalizationMethod canonicalizationMethod =
			signatureFactory.newCanonicalizationMethod(CanonicalizationMethod.EXCLUSIVE,
					(C14NMethodParameterSpec)null);
		SignatureMethod signatureMethod =
			signatureFactory.newSignatureMethod(signatureAlgorithm, null);
		SignedInfo signedInfo =
			signatureFactory.newSignedInfo(canonicalizationMethod,
				signatureMethod,
				Collections.singletonList(reference));
		return signedInfo;
	}

	/**
	 * Create a reference for the signature
	 * @param signatureFactory
	 * @param nodeId the ID of the node to sign
	 * @return a reference to the data which is to be covered by the signature
	 * @throws NoSuchAlgorithmException
	 * @throws InvalidAlgorithmParameterException
	 */
	protected static Reference makeReferenceById(
			XMLSignatureFactory signatureFactory, String nodeId, boolean sha256mode)
			throws NoSuchAlgorithmException, InvalidAlgorithmParameterException
	{
		String digestAlgorithm = null;
		if (sha256mode) {
			digestAlgorithm = DigestMethod.SHA256;
		} else {
			digestAlgorithm = DigestMethod.SHA1;
		}

		// Get the methods to use for creating the reference to the data to sign
		// (SHA1 for the digest, Exclusive Canonicalization for the
		// normalization of the XML)
		DigestMethod signatureDigest =
			signatureFactory.newDigestMethod(digestAlgorithm, null);
		Transform transformMethod =
			signatureFactory.newTransform(CanonicalizationMethod.EXCLUSIVE,
					(TransformParameterSpec)null);

		// Which node do we want to sign, in which manner
		Reference reference =
			signatureFactory.newReference(nodeId, 
				signatureDigest,
				Collections.singletonList(transformMethod),
				null,
				null);
		return reference;
	}


}
