package nl.nictiz.digsig.eh;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.security.InvalidKeyException;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.SignatureException;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateEncodingException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Properties;
import java.util.TimeZone;
import java.util.UUID;

import javax.security.auth.x500.X500Principal;

import nl.nictiz.digsig.SmartCard;
import nl.nictiz.digsig.StringBasedSignature;

/**
 * An example how to create WS-Security signatures for AORTA Electronic
 * Signatures using only Strings and PKCS#11 (cryptographic smart-card)
 * operations.
 */
public class EHStringBasedSignature extends StringBasedSignature {

	/**
	 * Constants for the namespaces used.
	 */
	private static final String EH_NAMESPACE =
		"http://www.aortarelease.nl/805/";

	private static final String EH_SPECIFICATION_VERSION =
		"http://www.aortarelease.nl/805/emd/1";

	/**
	 * The ID/reference URI in the signature.
	 */
	private static final String CERTIFICATE_REFERENCE = "signing-cert";

	/**
	 * HL7 date format to use with AORTA Electronic Signature
	 * NOTE: Electronic Signature dates are in local time, and must be matching
	 * the HL7 document if the same date is stated there. This includes
     * timezone information, if specified in the HL7.
	 */
	private final static DateFormat DATE_FORMAT =
		new SimpleDateFormat("yyyyMMddHHmmss");

	/**
	 * @param args
	 * @throws IOException 
	 * @throws CertificateException 
	 * @throws NoSuchAlgorithmException 
	 * @throws KeyStoreException 
	 * @throws UnrecoverableKeyException 
	 * @throws InvalidKeyException 
	 * @throws SignatureException 
	 */
	public static void main(String[] args)
		throws IOException, KeyStoreException, NoSuchAlgorithmException,
			CertificateException, UnrecoverableKeyException,
			InvalidKeyException, SignatureException
	{

		// Get the configuration to use
		Properties properties = SmartCard.loadPkcsConfiguration();

		// For the electronic signature in AORTA SHA-256 is mandated
		boolean sha256mode = true;

		// Access the smartcard
		char[] pincode = SmartCard.readPassword();
		SmartCard smartCard = new SmartCard(pincode, properties);

		// key will be a reference to the key on the smartcard, not the actual
		// key itself. Other parts of the libraries can cope with this "key".
		PrivateKey key = smartCard.getNonRepudiationKey();
		X509Certificate certificate = smartCard.getNonRepudiationCertificate();

		if (key == null || certificate == null) {
			System.err.println("Smartcard does not contain a non-repudiation" +
					" certificate, it cannot be used for an electronic" +
					" signature.");
			System.exit(1);
		}

		// The root OID for these messages.
		String rootOID = "2.16.840.1.113883.2.4.6.1.6005465.12.1";
		// Pick a random nonce, to be used as messageId.
		// Make sure to match this with the IDs in your infrastructure, and
		// that these IDs are unique under the same root. Add system
		// identifiers under your root if necessary.
		UUID uuid = UUID.randomUUID();
		String messageId = uuid.toString();
		messageId = "0003000201"; // XXX fixed ID for demonstration purposes

		// Used as a reference to the token from the signature
		// NOTE: This must be a valid XML id, mind the messageId you're using
		String wssTokenId = "id_" + rootOID + "_" + messageId;

		String token =
			createToken(rootOID, messageId, wssTokenId, certificate);

		System.out.println("SignedData:\n" + token + "\n");

		// The token above is created in canonical format
		String canonicalToken = token;

		// Get the base64 encoded digest of the token
		String encodedDigest = digestToken(canonicalToken, sha256mode);

		System.err.println("Digested token: " + encodedDigest);

		// This digest is used in the SignedInfo
		String canonicalizedSignedInfo =
			createSignedInfo(wssTokenId, encodedDigest, sha256mode);
		System.out.println("SignedInfo:\n" + canonicalizedSignedInfo + "\n");

		// Make the signature over the SignedInfo, which should also be base64
		// encoded for use in the XML messages
		String encodedSignature =
			signSignedInfo(key, canonicalizedSignedInfo, sha256mode);
		System.out.println("Signature:\n" + encodedSignature + "\n");

		String binarySecurityToken = createBST(certificate);
		String keyInfo = createKeyInfo(certificate);
		System.out.println("BinarySecurityToken:\n" + binarySecurityToken + "\n");
		System.out.println("KeyInfo:\n" + keyInfo + "\n");
	}

	/**
	 * Create a token for the given message content
	 * @param rootOID the root OID of the message ID
	 * @param messageId the extension of the message ID
	 *   The combination messageID and rootOID must be unique for the entire
	 *   "GBZ"! See the AORTA HL7 specifications...
	 * @param patientId the patient ID (BSN) 
	 * @param wssTokenId a reference to place in the token. Used in the
	 *   SignedInfo to refer to the token that is signed. Must be unique in the
	 *   whole resulting HL7 message document (SOAP message).
	 * @param certificate the certificate of the signature, needed to fill the
	 *   signature information with the certificate identification () 
	 * @return returns the token in canonical format
	 * @throws UnsupportedEncodingException 
	 */
	private static String createToken(String rootOID, String messageId,
			String wssTokenId, X509Certificate certificate)
		throws UnsupportedEncodingException
	{

		String prescriptionDateFormatted;

		// hardcoded sample data that matches the AORTA examples
		prescriptionDateFormatted = "20100718151043";
		// Alternative date using the current time
//		Calendar prescriptionDate = Calendar.getInstance();
//		prescriptionDateFormatted =
//			DATE_FORMAT.format(prescriptionDate.getTime());

		// Certificate issuer, in RFC 2253 format
		X500Principal certificateIssuer = certificate.getIssuerX500Principal();
		// Certificate serial number, this is (explicitly) in decimal notation,
		// in accordance with XML-Signature / WS-S Signature specifications
		String certificateSerialNumber =
			certificate.getSerialNumber().toString(10);
		String[] subjectAltName = extractUziString(certificate);

		// Data to include in the signedData prescription
		String patientName = "J.M. Breed";
		String patientGender = "Man";
		String patientBirthdate = "19680816";
		String patientId = "012345672";
		String authorName = "Dr. Frans Rijtje";
		String authorUziNumber = subjectAltName[2];
		String authorAddress = "Waterstraaat 14, Nattelanden";
		String authorRolecode = subjectAltName[5];
		String authorRolecodeDisplayname = "Apotheekhoudend huisarts";  // FIXME Read from vocabulary!
		String medicationCode = "999999";
		String medicationCodeSystem = "2.16.840.1.113883.2.4.4.7";
		String medicationText = "Propanolol hcl tablet 10mg";
		String medicationRepeatNumber = "2";
		String medicationQuantity = "56";
		String medicationQuantityUnit = "1";
		String medicationQuantityUnitDisplayname = "stuks";
		String usage = "eerst 2 weken lang 2x daags 2 tabletten dan aansluitend 2 weken lang 2x daags 1 tablet en 4 weken lang 1x daags 1 tablet";

		// NOTE: the following String is in the XML canonical format. It
		// contains no newlines, nor whitespaces between elements. Quotes and
		// prefixes can <em>not</em> be altered once the digest is calculated.
		String token =
			"<signedDataPrescription xmlns=\"" + EH_NAMESPACE + "\" xmlns:wsu=\"" + WSU_NAMESPACE + "\" wsu:Id=\"" + wssTokenId + "\">"
				+ "<signatureMetaData>"
				+ "<signatureVersion>" + EH_SPECIFICATION_VERSION + "</signatureVersion>"
				// The namespace is declared explicitly, since the signedData element
				// does not contain any other element in the same namespace. This to
				// make sure the token is in proper exclusive-C14N notation.
				+ "<ds:X509IssuerSerial xmlns:ds=\"" + XML_DSIG_NAMESPACE + "\">"
				+ "<ds:X509IssuerName>" + certificateIssuer + "</ds:X509IssuerName>"
				+ "<ds:X509SerialNumber>" + certificateSerialNumber + "</ds:X509SerialNumber>"
				+ "</ds:X509IssuerSerial>"
				+ "</signatureMetaData>"
				+ "<prescription>"
				+ "<id>" // prescription itself
				+ "<root>" + rootOID + "</root>"
				+ "<extension>" + messageId + "</extension>"
				+ "</id>"
				+ "<dateTime>" + prescriptionDateFormatted + "</dateTime>"
				+ "<patient>"
				+ "<id>" // patient
				+ "<root>2.16.840.1.113883.2.4.6.3</root>"   // BSN vocabulary
				+ "<extension>" + patientId + "</extension>"
				+ "</id>"
				+ "<name>" + patientName + "</name>"
				+ "<gender>" + patientGender + "</gender>"
				+ "<birthdate>" + patientBirthdate + "</birthdate>"
				+ "</patient>"
				+ "<author>"
				+ "<id>" //author
				+ "<root>2.16.528.1.1007.3.1</root>" // UZI personal number vocabulary
				+ "<extension>" + authorUziNumber + "</extension>"
				+ "</id>"
				+ "<rolcode>"
				+ "<codeSystem>2.16.840.1.113883.2.4.15.111</codeSystem>" // UZI rolcode vocabulary
				+ "<code>" + authorRolecode + "</code>"
				+ "<displayName>" + authorRolecodeDisplayname + "</displayName>"
				+ "</rolcode>"
				+ "<name>" + authorName + "</name>"
				+ "<address>" + authorAddress + "</address>"
				+ "</author>"
				+ "<medication>"
				+ "<codeSystem>" + medicationCodeSystem + "</codeSystem>"
				+ "<code>" + medicationCode + "</code>"
				+ "<displayName>" + medicationText + "</displayName>"
				+ "</medication>"
				+ "<quantity>"
				+ "<iter>" + medicationRepeatNumber + "</iter>"
				+ "<value>" + medicationQuantity + "</value>"
				+ "<unit>"
				+ "<code>" + medicationQuantityUnit + "</code>"
				+ "<displayName>" + medicationQuantityUnitDisplayname + "</displayName>"
				+ "</unit>"
				+ "</quantity>"
				+ "<usage>" + usage + "</usage>"
				+ "</prescription>"
				+ "</signedDataPrescription>";

		return token;
	}

	/**
	 * Extract the "UZI-string" from the certificate. It will read and parse the
	 * String with the CA, version, UZI-number, URA and rolecode from a
	 * certificate.
	 * XXX Note that it does not handle errors properly and is error prone. Use
	 * a ASN.1 library for proper parsing!
	 * @param certificate
	 * @return
	 * @throws UnsupportedEncodingException
	 */
	private static String[] extractUziString(X509Certificate certificate)
			throws UnsupportedEncodingException {

		// Get the subject alternative name (well known OID).
		byte[] extensionValue = certificate.getExtensionValue("2.5.29.17");
		// Extract the "UZI-string" from the subjectAlternativeName
		// XXX not recommended, counting bytes, use ASN.1 library instead!
		String subjectAlternativeName =
			new String(extensionValue, 14, extensionValue.length - 14, "UTF-8");
		String[] subjectAltName = subjectAlternativeName.split("-");
		return subjectAltName;
	}

	/**
	 * Create the WS-S BinarySecurityToken for the given certificate.
	 * @param certificate the certificate used to sign the message. 
	 * @return the WS-S BinarySecurityToken for the certificate
	 */
	private static String createBST(X509Certificate certificate)
		throws CertificateEncodingException {
		String result = "<wss:BinarySecurityToken wsu:Id=\"" + CERTIFICATE_REFERENCE + "\" ValueType=\"http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-x509-token-profile-1.0#X509v3\" EncodingType=\"http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-soap-message-security-1.0#Base64Binary\">"
			+ base64encode(certificate.getEncoded())
			+ "</wss:BinarySecurityToken>";
		return result;
	}

	/**
	 * Create the WS-S KeyInfo for the given certificate.
	 * @param certificate the certificate used to sign the message. 
	 * @return the KeyInfo for the certificate
	 */
	protected static String createKeyInfo(X509Certificate certificate) {
		// The URI must match the wsu:Id in the BST. This matching is relative
		// in the document, so the URI is the ID with a # prepended (like with
		// same document HTML <A>-anchors).
		String result = "<KeyInfo>"
			+ "<wss:SecurityTokenReference>"
			+ "<wss:Reference URI=\"#" + CERTIFICATE_REFERENCE + "\" ValueType=\"http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-x509-token-profile-1.0#X509v3\" />"
			+ "</wss:SecurityTokenReference>"
			+ "</KeyInfo>";
		return result;
	}

}
