package nl.nictiz.digsig.eh;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.cert.CertificateException;
import java.util.Properties;
import java.util.Vector;

import javax.xml.crypto.dsig.DigestMethod;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactoryConfigurationError;

import nl.nictiz.digsig.SmartCard;
import nl.nictiz.digsig.WssCommon;
import nl.nictiz.digsig.XmlCommon;

import org.apache.ws.security.WSConstants;
import org.apache.ws.security.WSEncryptionPart;
import org.apache.ws.security.components.crypto.Crypto;
import org.apache.ws.security.message.WSSecHeader;
import org.apache.ws.security.message.WSSecSignature;
import org.apache.xml.security.signature.XMLSignature;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;

/**
 * Example of creating the WS-Security signature with a WS-S library.
 */
public class EHWssLibrarySignature
	extends WssCommon
{

	/**
	 * @param args
	 * @throws IOException 
	 * @throws SAXException 
	 * @throws ParserConfigurationException 
	 * @throws FileNotFoundException 
	 * @throws CertificateException 
	 * @throws NoSuchAlgorithmException 
	 * @throws KeyStoreException 
	 * @throws TransformerException 
	 * @throws TransformerFactoryConfigurationError 
	 * @throws TransformerConfigurationException 
	 */
	public static void main(String[] args)
		throws FileNotFoundException, ParserConfigurationException,
		SAXException, IOException, KeyStoreException, NoSuchAlgorithmException,
		CertificateException, TransformerFactoryConfigurationError, TransformerException
	{

		// For the electronic signature in AORTA SHA-256 is mandated
		boolean sha256mode = true;

		// load the document to sign
		String filename = args[0];
		Document document = XmlCommon.getDocument(filename);

		// Get the configuration to use
		Properties pkcs11properties = SmartCard.loadPkcsConfiguration();

		// Access the smartcard
		char[] pincode = SmartCard.readPassword();
		SmartCard smartCard = new SmartCard(pincode, pkcs11properties);

		if (smartCard.getNonRepudiationCertificateAlias() == null) {
			System.err.println("Smartcard does not contain a non-repudiation" +
					" certificate, it cannot be used for an electronic" +
					" signature.");
			System.exit(1);
		}
		
		Crypto crypto = makeWssCrypto();

		// Create a WS-Security SOAP-header in the document
		WSSecHeader header = new WSSecHeader();
		header.setMustUnderstand(true);
		header.setActor("http://www.aortarelease.nl/actor/gbx");
		header.insertSecurityHeader(document);

		// Create a WS-Security signature
		WSSecSignature signature = new WSSecSignature();

		if (sha256mode) {
			signature.setDigestAlgo(DigestMethod.SHA256);
			signature.setSignatureAlgorithm(XMLSignature.ALGO_ID_SIGNATURE_RSA_SHA256);
		}
		
		// We use the BinarySecurityToken (BST) way to identify the user
		signature.setKeyIdentifierType(WSConstants.BST_DIRECT_REFERENCE);

		// Use Exclusive XML canonicalization, without comments
		signature.setSigCanonicalization(WSConstants.C14N_EXCL_OMIT_COMMENTS);

		// The alias is a reference to the certificate in the Java KeyStore
		signature.setUserInfo(smartCard.getNonRepudiationCertificateAlias(), "");

		// Identify the part of the message to sign.
		// This part references the whole "wsu:Id" element in the message. This
		// identifier should already be part of the message, identifying the
		// token.
		// Note: the library supports both signatures and encryption. Both
		// work through a reference, the name does imply encryption,
		// so this is a bit of a misnomer in WSS4J.
		Vector<WSEncryptionPart> references = new Vector<WSEncryptionPart>();
		WSEncryptionPart part =
			new WSEncryptionPart("id_2.16.840.1.113883.2.4.6.1.6005465.12.1_0003000201",
					"Element");
		references.add(part);
		signature.setParts(references);

		// Compute the signature
		signature.build(document, crypto, header);

		// Add it to the SOAP-header
		signature.prependToHeader(header);

		// Show the resulting document
		XmlCommon.displayXml(document);

	}

}
