package nl.nictiz.digsig.eh;

import java.io.IOException;
import java.security.InvalidAlgorithmParameterException;
import java.security.Key;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.Collections;
import java.util.Properties;

import javax.xml.crypto.MarshalException;
import javax.xml.crypto.dsig.SignedInfo;
import javax.xml.crypto.dsig.XMLSignature;
import javax.xml.crypto.dsig.XMLSignatureException;
import javax.xml.crypto.dsig.XMLSignatureFactory;
import javax.xml.crypto.dsig.dom.DOMSignContext;
import javax.xml.crypto.dsig.keyinfo.KeyInfo;
import javax.xml.crypto.dsig.keyinfo.KeyInfoFactory;
import javax.xml.crypto.dsig.keyinfo.X509Data;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;

import nl.nictiz.digsig.SmartCard;
import nl.nictiz.digsig.XmlCommon;
import nl.nictiz.digsig.XmlLibrarySignature;

import org.w3c.dom.Document;
import org.xml.sax.SAXException;

/**
 * An example how to create WS-Security signatures using a XML-Signature
 * library.
 */
public class EHXmlLibrarySignature extends XmlLibrarySignature {

	/**
	 * @param args
	 * @throws CertificateException 
	 * @throws NoSuchAlgorithmException 
	 * @throws KeyStoreException 
	 * @throws UnrecoverableKeyException 
	 * @throws SAXException 
	 * @throws ParserConfigurationException 
	 * @throws TransformerException 
	 * @throws XMLSignatureException 
	 * @throws MarshalException 
	 * @throws InvalidAlgorithmParameterException 
	 */
	public static void main(String[] args)
		throws IOException, KeyStoreException, NoSuchAlgorithmException,
			CertificateException, UnrecoverableKeyException,
			ParserConfigurationException, SAXException,
			InvalidAlgorithmParameterException, MarshalException,
			XMLSignatureException, TransformerException
	{

		// For the electronic signature in AORTA SHA-256 is mandated
		boolean sha256mode = true;

		// Which document to sign?
		String filename = args[0];
		Document document = XmlCommon.getDocument(filename);

		// Get the configuration to use
		Properties properties = SmartCard.loadPkcsConfiguration();

		// Access the smartcard
		char[] pincode = SmartCard.readPassword();
		SmartCard smartCard = new SmartCard(pincode, properties);

		// key will be a reference to the key on the smartcard, not the actual
		// key itself. Other parts of the libraries can cope with this "key".
		PrivateKey key = smartCard.getNonRepudiationKey();
		X509Certificate certificate = smartCard.getNonRepudiationCertificate();

		if (key == null || certificate == null) {
			System.err.println("Smartcard does not contain a non-repudiation" +
					" certificate, it cannot be used for an electronic" +
					" signature.");
			System.exit(1);
		}
		
		// Add the signature to the document
		createSignature(document, key, certificate,
				"#id_2.16.840.1.113883.2.4.6.1.6005465.12.1_0003000201", sha256mode);

		// NOTE KeyInfo in WS-Security differs from plain XML-Signature: The
		// certificate is included in a separate BinarySecurityToken element,
		// which is referenced from the KeyInfo. XML-Signature places the X509
		// certificate in the KeyInfo. Since the KeyInfo itself is not signed,
		// the signature can be transformed.
		// Both the WSS4J and Metro implementations of WSS use internals of
		// their related XML-Security library to accomplish this. A pure
		// java.xml.crypto solution does not seem possible, without
		// implementing a custom KeyInfoFactory :-(
		Document wssDocument =
			XmlCommon.transform(document, "eh-xmlsignature-to-wss.xsl");

		// display the result
		XmlCommon.displayXml(wssDocument);

	}


	/**
	 * Create a XML-signature
	 * @param document the document to contain the signature
	 * @param key the key for the signature
	 * @param certificate the X509certificate that identifies the owner of
	 *     the key (i.e. smart-card)
	 * @param nodeId the ID of the node to sign in the document
	 * @throws NoSuchAlgorithmException
	 * @throws InvalidAlgorithmParameterException
	 * @throws MarshalException
	 * @throws XMLSignatureException
	 * @throws TransformerException 
	 * @throws IOException 
	 */
	protected static void createSignature(Document document, Key key,
			X509Certificate certificate, String nodeId, boolean sha256mode)
	    throws NoSuchAlgorithmException, InvalidAlgorithmParameterException,
	    	MarshalException, XMLSignatureException, TransformerException,
	    	IOException
	{
		// Some Java boilerplate
		XMLSignatureFactory signatureFactory =
			XMLSignatureFactory.getInstance("DOM");

		// The information about the signature itself
		SignedInfo signedInfo =
			makeSignedInfo(signatureFactory, nodeId, sha256mode);

		// The information about the signer
		KeyInfo keyInfo = makeKeyInfo(certificate);

		// Create the actual signature
		DOMSignContext signContext =
			new DOMSignContext(key, document.getDocumentElement());
		XMLSignature signature =
			signatureFactory.newXMLSignature(signedInfo, keyInfo);
		signature.sign(signContext);

//		showCanonicalizedForm(signedInfo);

	}

	/**
	 * Create the information about the key used to create the signature
	 * @param certificate
	 * @return
	 */
	protected static KeyInfo makeKeyInfo(X509Certificate certificate) {

		KeyInfoFactory keyInfoFactory = KeyInfoFactory.getInstance();

		// AORTA's "elektronische handtekening" includes the entire certificate,
		// not just a reference
		X509Data x509Data =
			keyInfoFactory.newX509Data(Collections.singletonList(certificate));

		KeyInfo keyInfo =
			keyInfoFactory.newKeyInfo(Collections.singletonList(x509Data));

		return keyInfo;
	}

}
