package nl.nictiz.digsig.ta;

import java.io.IOException;
import java.security.InvalidKeyException;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.SignatureException;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Calendar;
import java.util.List;
import java.util.Properties;
import java.util.TimeZone;
import java.util.UUID;

import javax.security.auth.x500.X500Principal;

import nl.nictiz.digsig.SmartCard;
import nl.nictiz.digsig.StringBasedSignature;

/**
 * An example how to create WS-Security signatures for Token Authentication
 * using only Strings and PKCS#11 (cryptographic smart-card) operations.
 */
public class TAStringBasedSignature extends StringBasedSignature {

	/**
	 * Constant for the namespace used in the token.
	 */
	private static final String TA_NAMESPACE =
		"http://www.aortarelease.nl/805/";

	/**
	 * HL7 date format to use with Token Authentication
	 * NOTE: Must be in UTC for Token Authentication!
	 */
	private final static DateFormat DATE_FORMAT =
		new SimpleDateFormat("yyyyMMddHHmmss");
	static {
		// Explicitly set the date formatter to make UTC times
		DATE_FORMAT.setTimeZone(TimeZone.getTimeZone("UTC"));
	}

	/**
	 * @param args
	 * @throws IOException 
	 * @throws CertificateException 
	 * @throws NoSuchAlgorithmException 
	 * @throws KeyStoreException 
	 * @throws UnrecoverableKeyException 
	 * @throws InvalidKeyException 
	 * @throws SignatureException 
	 */
	public static void main(String[] args)
		throws IOException, KeyStoreException, NoSuchAlgorithmException,
			CertificateException, UnrecoverableKeyException,
			InvalidKeyException, SignatureException
	{
		// switch SHA-256 mode based on program startup argument
		boolean sha256mode = false;
		List<String> arguments = Arrays.asList(args);
		if (arguments.contains("-sha256")) {
			sha256mode = true;
		}

		// Get the configuration to use
		Properties properties = SmartCard.loadPkcsConfiguration();

		// Access the smartcard
		char[] pincode = SmartCard.readPassword();
		SmartCard smartCard = new SmartCard(pincode, properties);

		// key will be a reference to the key on the smartcard, not the actual
		// key itself. Other parts of the libraries can cope with this "key".
		PrivateKey key = smartCard.getAuthenticationKey();
		X509Certificate certificate = smartCard.getAuthenticationCertificate();

		// Which trigger
		String trigger = "QURX_TE990011NL";
		// Which patient
		String patientId = "012345672";

		// The root OID for these messages.
		String rootOID = "2.16.528.1.1007.3.3.1234567.1";
		// Pick a random nonce, to be used as messageId.
		// Make sure to match this with the IDs in your infrastructure, and
		// that these IDs are unique under the same root. Add system
		// identifiers under your root if necessary.
		UUID uuid = UUID.randomUUID();
		String messageId = uuid.toString();
		messageId = "0123456789"; // XXX fixed ID for demonstration purposes

		// Used as a reference to the token from the signature
		// NOTE: This must be a valid XML id, mind the messageId you're using
		String wssTokenId = "token_" + rootOID + "_" + messageId;

		String token =
			createToken(rootOID, messageId, trigger, patientId, wssTokenId);

		System.out.println("SignedData\n" + token + "\n");
		
		// The token above is created in canonical format
		String canonicalToken = token;

		// Get the base64 encoded digest of the token
		String encodedDigest = digestToken(canonicalToken, sha256mode);

		System.err.println("Digested token: " + encodedDigest);

		// This digest is used in the SignedInfo
		String canonicalizedSignedInfo =
			createSignedInfo(wssTokenId, encodedDigest, sha256mode);
		System.out.println("SignedInfo\n" + canonicalizedSignedInfo + "\n");

		// Make the signature over the SignedInfo, which should also be base64
		// encoded for use in the XML messages
		String encodedSignature =
			signSignedInfo(key, canonicalizedSignedInfo, sha256mode);
		System.out.println("Signature:\n" + encodedSignature + "\n");

		String keyInfo = createKeyInfo(certificate);

		System.out.println("KeyInfo:\n" + keyInfo + "\n");
		
	}

	/**
	 * Create a token for the given message content
	 * @param rootOID the root OID of the message ID
	 * @param messageId the extension of the message ID
	 *   The combination messageID and rootOID must be unique for the entire
	 *   "GBZ"! See the AORTA HL7 specifications...
	 * @param trigger the triggerEventId for the message 
	 * @param patientId the patient ID (BSN) 
	 * @param wssTokenId a reference to place in the token. Used in the
	 *   SignedInfo to refer to the token that is signed. Must be unique in the
	 *   whole resulting HL7 message document (SOAP message).
	 * @return returns the token in canonical format
	 */
	private static String createToken(String rootOID, String messageId,
			String trigger, String patientId, String wssTokenId)
	{

		String validFromFormatted;
		String validUntilFormatted;

		// hardcoded sample data that matches the AORTA examples
		validFromFormatted = "20050128173600";
		validUntilFormatted = "20050128174059";
		// Alternative data using the current time
//		// Validity default now plus 1 hour
//		Calendar validFrom = Calendar.getInstance();
//		Calendar validUntil = (Calendar) validFrom.clone();
//		validUntil.setTime(validFrom.getTime());
//		validUntil.add(Calendar.HOUR, 1);
//		validFromFormatted = DATE_FORMAT.format(validFrom.getTime());
//		validUntilFormatted = DATE_FORMAT.format(validUntil.getTime());

		// NOTE: the following String is in the XML canonical format. It
		// contains no newlines, nor whitespaces between elements. Quotes and
		// prefixes can <em>not</em> be altered once the digest is calculated.
		String token =
			"<signedData xmlns=\"" + TA_NAMESPACE + "\" xmlns:wsu=\"" + WSU_NAMESPACE + "\" wsu:Id=\"" + wssTokenId + "\">"
		    + "<authenticationData>"
		    + "<messageId>"
		    + "<root>" + rootOID + "</root>"
		    + "<extension>" + messageId + "</extension>"
		    + "</messageId>"
		    + "<notBefore>" + validFromFormatted + "</notBefore>"
		    + "<notAfter>" + validUntilFormatted + "</notAfter>"
		    + "<addressedParty>"
		    + "<root>2.16.840.1.113883.2.4.6.6</root>"   // LSP root
		    + "<extension>1</extension>"                 // LSP id
		    + "</addressedParty>"
		    + "</authenticationData>"
		    + "<coSignedData>"
		    + "<triggerEventId>" + trigger + "</triggerEventId>"
		    + "<patientId>"
		    + "<root>2.16.840.1.113883.2.4.6.3</root>"   // BSN vocabulary
		    + "<extension>" + patientId + "</extension>" // patient BSN
		    + "</patientId>"
		    + "</coSignedData>"
		    + "</signedData>";
		return token;
	}

	/**
	 * Create the WS-S KeyInfo for the given certificate.
	 * @param certificate the certificate used to sign the message. 
	 * @return the KeyInfo for the certificate
	 */
	private static String createKeyInfo(X509Certificate certificate) {
		// Certificate issuer, in RFC 2253 format
		X500Principal certificateIssuer = certificate.getIssuerX500Principal();
		// Certificate serial number, this is (explicitly) in decimal notation,
		// in accordance with XML-Signature / WS-S Signature specifications
		String certificateSerialNumber =
			certificate.getSerialNumber().toString(10);
		String result = "<KeyInfo>"
			+ "<wss:SecurityTokenReference>"
			+ "<ds:X509Data xmlns:ds=\"" + XML_DSIG_NAMESPACE + "\">"
			+ "<ds:X509IssuerSerial>"
			+ "<ds:X509IssuerName>" + certificateIssuer + "</ds:X509IssuerName>"
			+ "<ds:X509SerialNumber>" + certificateSerialNumber + "</ds:X509SerialNumber>"
			+ "</ds:X509IssuerSerial>"
			+ "</ds:X509Data>"
			+ "</wss:SecurityTokenReference>"
			+ "</KeyInfo>";
		return result;
	}


}
